
/*
 * slotsig_base.h header file
 *
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 * This file provides base support for signals and slots
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#ifndef __SLOTSIG_BASES_H__
#define __SLOTSIG_BASES_H__ 1

#include <map>

namespace SlotSig {

class SlotBase
{
  public:
    virtual ~SlotBase() { }
} ; // SlotBase

class SlotsSetBase ;

class SignalBase
{
  friend class SlotsSetBase ;
  typedef std::multimap<SlotsSetBase*, SlotBase*> slots_sets_map ;
  typedef slots_sets_map::const_iterator slots_sets_const_iterator ;
  typedef slots_sets_map::iterator slots_sets_iterator ;
  slots_sets_map slots_sets ;

  public:
    SignalBase() :
      slots_sets()
    { }

    virtual ~SignalBase() ;

  protected:
    virtual void disconnectSlot(SlotBase*& sl) = 0 ;

    template <class _C>
    void connectTo (_C* c, SlotBase* sl)
    {
      SlotsSetBase* ssb = dynamic_cast<SlotsSetBase*>(c) ;
      if ( (ssb == 0) || (sl == 0) )
        return ;
      ssb->connectFrom(this) ;
      slots_sets.insert(std::make_pair(ssb, sl)) ;
    }

    template <class _C>
    void disconnectTo (_C* c, SlotBase* sl)
    {
      SlotsSetBase* ssb = dynamic_cast<SlotsSetBase*>(c) ;
      if ( (ssb == 0) || (sl == 0) )
        return ;
      std::pair<slots_sets_iterator, slots_sets_iterator> slots =
        slots_sets.equal_range(ssb) ;
      if ( slots.first == slots.second )
        return ;
      slots_sets_iterator iter = slots.first ;
      while ( (iter != slots.second) && (iter->second != sl) )
        ++iter ;
      if ( iter->second == sl )
      {
        disconnectSlot(iter->second) ;
        slots_sets.erase(iter) ;
      }
      ssb->disconnectFrom(this) ;
    }

  private:
    void slotsSetDeleted (SlotsSetBase* ssb)
    {
      if ( ssb == 0 )
        return ;
      std::pair<slots_sets_iterator, slots_sets_iterator> slots =
        slots_sets.equal_range(ssb) ;
      if ( slots.first == slots.second )
        return ;
      slots_sets_iterator iter = slots.first ;
      while ( iter != slots.second )
        disconnectSlot((iter++)->second) ;
      slots_sets.erase(slots.first, slots.second) ;
    }
} ; // class SignalBase

class SlotsSetBase
{
  friend class SignalBase ;
  typedef std::map<SignalBase*, unsigned> signals_map ;
  signals_map signals ;

  public:
    SlotsSetBase() :
      signals()
    { }

    virtual ~SlotsSetBase()
    {
      const signals_map::const_iterator end = signals.end() ;
      signals_map::const_iterator iter = signals.begin() ;
      while ( iter != end )
        (iter++)->first->slotsSetDeleted(this) ;
    }

  private:
    void connectFrom(SignalBase* sig)
    {
      if ( sig == 0 )
        return ;
      ++(signals[sig]) ;
    }

    void disconnectFrom(SignalBase* sig)
    {
      if ( sig == 0 )
        return ;
      signals_map::iterator iter = signals.find(sig) ;
      if ( iter == signals.end() )
        return ;
      if ( iter->second == 1 )
        signals.erase(iter) ;
      else
        --(iter->second) ;
    }

    void signalDeleted(SignalBase* sig)
    {
      if ( sig == 0 )
        return ;
      signals.erase(sig) ;
    }
} ; // class SlotsSetBase

SignalBase::~SignalBase()
{
  const slots_sets_map::const_iterator end = slots_sets.end() ;
  slots_sets_map::const_iterator iter = slots_sets.begin() ;
  while ( iter != end )
  {
    SlotsSetBase* ssb = iter->first ;
    ssb->signalDeleted(this) ;
    ++iter ;
  }
}

} // namespace SlotSig

#endif // __SLOTSIG_BASES_H__
